import json
import uuid
import logging
from typing import Optional, List, Callable
from .client import authorize

logger = logging.getLogger("payfence")


class PayFenceMiddleware:
    """ASGI middleware for PayFence authorization."""

    def __init__(
        self,
        app,
        api_key: str,
        base_url: str = "https://api.payfence.io",
        exclude_paths: Optional[List[str]] = None,
        on_deny: Optional[Callable] = None,
    ):
        self.app = app
        self.api_key = api_key
        self.base_url = base_url
        self.exclude_paths = exclude_paths or []
        self.on_deny = on_deny

    async def __call__(self, scope, receive, send):
        if scope["type"] != "http":
            return await self.app(scope, receive, send)

        path = scope.get("path", "/")

        # Skip excluded paths
        if any(path.startswith(p) for p in self.exclude_paths):
            logger.debug("Path excluded: %s", path)
            return await self.app(scope, receive, send)

        # Extract bearer token from headers
        headers = dict(scope.get("headers", []))
        auth = headers.get(b"authorization", b"").decode()
        token = auth[7:] if auth.startswith("Bearer ") else ""

        if not token:
            logger.info("Deny: missing Bearer token for %s", path)
            return await self._send_deny(send, {"decision": "deny", "reason": "payment_required", "message": "Missing Bearer token"})

        try:
            request_id = headers.get(b"x-request-id", b"").decode() or str(uuid.uuid4())
            method = scope.get("method", "GET")

            decision = authorize(
                api_key=self.api_key,
                token=token,
                method=method,
                path=path,
                request_id=request_id,
                base_url=self.base_url,
            )

            if decision.get("decision") == "allow":
                logger.debug("Allow: %s %s", method, path)
                scope["payfence"] = decision
                return await self.app(scope, receive, send)

            # Deny
            logger.info("Deny: %s %s — reason=%s", method, path, decision.get("reason"))
            if self.on_deny:
                return await self.on_deny(scope, receive, send, decision)
            return await self._send_deny(send, decision)

        except Exception as e:
            logger.warning("PayFence authorize failed, failing open: %s", e)
            return await self.app(scope, receive, send)

    async def _send_deny(self, send, decision):
        body = json.dumps(decision).encode()
        await send({
            "type": "http.response.start",
            "status": 402,
            "headers": [[b"content-type", b"application/json"]],
        })
        await send({
            "type": "http.response.body",
            "body": body,
        })



import hashlib
import hmac
import time

from payfence.hmac_verify import verify_signature

SECRET = "test-secret-key"


def _sign(method, path, timestamp, request_id, body):
    body_hash = hashlib.sha256(body.encode() if isinstance(body, str) else body).hexdigest()
    canonical = f"{method}\n{path}\n{timestamp}\n{request_id}\n{body_hash}"
    return hmac.new(SECRET.encode(), canonical.encode(), hashlib.sha256).hexdigest()


def _now():
    return str(int(time.time()))


class TestVerifySignature:
    def test_valid_signature(self):
        ts = _now()
        sig = _sign("POST", "/v1/weather", ts, "req-1", '{"city":"NYC"}')
        assert verify_signature(
            secret=SECRET, method="POST", path="/v1/weather",
            timestamp=ts, request_id="req-1", body='{"city":"NYC"}', signature=sig,
        ) is True

    def test_wrong_signature(self):
        ts = _now()
        assert verify_signature(
            secret=SECRET, method="POST", path="/v1/weather",
            timestamp=ts, request_id="req-1", body='{"city":"NYC"}', signature="bad",
        ) is False

    def test_expired_timestamp(self):
        ts = str(int(time.time()) - 600)
        sig = _sign("POST", "/v1/weather", ts, "req-1", '{"city":"NYC"}')
        assert verify_signature(
            secret=SECRET, method="POST", path="/v1/weather",
            timestamp=ts, request_id="req-1", body='{"city":"NYC"}', signature=sig,
        ) is False

    def test_invalid_timestamp(self):
        assert verify_signature(
            secret=SECRET, method="POST", path="/v1/weather",
            timestamp="not-a-number", request_id="req-1", body="", signature="x",
        ) is False

    def test_custom_tolerance(self):
        ts = str(int(time.time()) - 10)
        sig = _sign("GET", "/v1/data", ts, "req-2", "")
        # 5s tolerance → fail
        assert verify_signature(
            secret=SECRET, method="GET", path="/v1/data",
            timestamp=ts, request_id="req-2", body="", signature=sig,
            tolerance_seconds=5,
        ) is False
        # 15s tolerance → pass
        assert verify_signature(
            secret=SECRET, method="GET", path="/v1/data",
            timestamp=ts, request_id="req-2", body="", signature=sig,
            tolerance_seconds=15,
        ) is True

    def test_bytes_body(self):
        ts = _now()
        body = b'{"raw":true}'
        sig = _sign("POST", "/v1/data", ts, "req-3", body.decode())
        assert verify_signature(
            secret=SECRET, method="POST", path="/v1/data",
            timestamp=ts, request_id="req-3", body=body, signature=sig,
        ) is True

    def test_empty_signature(self):
        ts = _now()
        assert verify_signature(
            secret=SECRET, method="GET", path="/",
            timestamp=ts, request_id="req-4", body="", signature="",
        ) is False



import json
import asyncio
from unittest.mock import patch, AsyncMock, MagicMock

from payfence.middleware import PayFenceMiddleware


def run(coro):
    return asyncio.get_event_loop().run_until_complete(coro)


def make_scope(path="/v1/weather", method="GET", token="tok_test_123", headers=None):
    hdrs = headers if headers is not None else [
        (b"authorization", f"Bearer {token}".encode()),
    ]
    return {"type": "http", "path": path, "method": method, "headers": hdrs}


def capture_response():
    """Returns (send_fn, messages_list) to capture ASGI send() calls."""
    messages = []

    async def send(msg):
        messages.append(msg)

    return send, messages


class TestPayFenceMiddleware:
    @patch("payfence.middleware.authorize")
    def test_allow_passes_to_app(self, mock_auth):
        mock_auth.return_value = {"decision": "allow", "units": 1}
        app = AsyncMock()

        mw = PayFenceMiddleware(app, api_key="sk_test")
        scope = make_scope()
        run(mw(scope, AsyncMock(), AsyncMock()))

        app.assert_called_once()
        assert scope["payfence"]["decision"] == "allow"

    @patch("payfence.middleware.authorize")
    def test_deny_returns_402(self, mock_auth):
        mock_auth.return_value = {"decision": "deny", "reason": "payment_required"}
        app = AsyncMock()

        mw = PayFenceMiddleware(app, api_key="sk_test")
        send, messages = capture_response()
        run(mw(make_scope(), AsyncMock(), send))

        app.assert_not_called()
        assert messages[0]["status"] == 402
        body = json.loads(messages[1]["body"])
        assert body["decision"] == "deny"

    @patch("payfence.middleware.authorize")
    def test_missing_token_returns_402(self, mock_auth):
        app = AsyncMock()
        mw = PayFenceMiddleware(app, api_key="sk_test")
        scope = make_scope(headers=[])  # no auth header
        send, messages = capture_response()

        run(mw(scope, AsyncMock(), send))

        mock_auth.assert_not_called()
        assert messages[0]["status"] == 402
        body = json.loads(messages[1]["body"])
        assert body["reason"] == "payment_required"

    @patch("payfence.middleware.authorize")
    def test_excluded_path_skips_auth(self, mock_auth):
        app = AsyncMock()
        mw = PayFenceMiddleware(app, api_key="sk_test", exclude_paths=["/health"])
        scope = make_scope(path="/health")

        run(mw(scope, AsyncMock(), AsyncMock()))

        mock_auth.assert_not_called()
        app.assert_called_once()

    @patch("payfence.middleware.authorize")
    def test_network_error_fails_open(self, mock_auth):
        mock_auth.side_effect = ConnectionError("ECONNREFUSED")
        app = AsyncMock()

        mw = PayFenceMiddleware(app, api_key="sk_test")
        run(mw(make_scope(), AsyncMock(), AsyncMock()))

        app.assert_called_once()

    @patch("payfence.middleware.authorize")
    def test_on_deny_callback(self, mock_auth):
        mock_auth.return_value = {"decision": "deny", "reason": "quota_exceeded"}
        app = AsyncMock()
        on_deny = AsyncMock()

        mw = PayFenceMiddleware(app, api_key="sk_test", on_deny=on_deny)
        run(mw(make_scope(), AsyncMock(), AsyncMock()))

        on_deny.assert_called_once()
        assert on_deny.call_args[0][3]["reason"] == "quota_exceeded"
        app.assert_not_called()

    def test_non_http_scope_passes_through(self):
        app = AsyncMock()
        mw = PayFenceMiddleware(app, api_key="sk_test")
        scope = {"type": "websocket", "path": "/ws"}

        run(mw(scope, AsyncMock(), AsyncMock()))

        app.assert_called_once()

    @patch("payfence.middleware.authorize")
    def test_custom_base_url_forwarded(self, mock_auth):
        mock_auth.return_value = {"decision": "allow"}
        app = AsyncMock()

        mw = PayFenceMiddleware(app, api_key="sk_test", base_url="https://custom.io")
        run(mw(make_scope(), AsyncMock(), AsyncMock()))

        _, kwargs = mock_auth.call_args
        assert kwargs["base_url"] == "https://custom.io"



import hashlib
import hmac
import logging
import time
from typing import Union

logger = logging.getLogger("payfence")


def verify_signature(
    secret: str,
    method: str,
    path: str,
    timestamp: str,
    request_id: str,
    body: Union[str, bytes],
    signature: str,
    tolerance_seconds: int = 300,
) -> bool:
    """Verify a PayFence HMAC signature from proxy mode."""
    try:
        ts = int(timestamp)
    except (ValueError, TypeError):
        logger.info("HMAC rejected: invalid timestamp '%s'", timestamp)
        return False

    age = abs(time.time() - ts)
    if age > tolerance_seconds:
        logger.info("HMAC rejected: timestamp expired (age=%ds, tolerance=%ds)", int(age), tolerance_seconds)
        return False

    if isinstance(body, str):
        body = body.encode()

    body_hash = hashlib.sha256(body).hexdigest()
    canonical = f"{method}\n{path}\n{timestamp}\n{request_id}\n{body_hash}"
    expected = hmac.new(secret.encode(), canonical.encode(), hashlib.sha256).hexdigest()

    valid = hmac.compare_digest(signature, expected)
    if valid:
        logger.debug("HMAC verified: %s %s", method, path)
    else:
        logger.info("HMAC rejected: signature mismatch for %s %s", method, path)
    return valid



import json
from unittest.mock import patch, MagicMock
from urllib.error import HTTPError
from io import BytesIO

from payfence.client import authorize


class TestAuthorize:
    @patch("payfence.client.urlopen")
    def test_allow_response(self, mock_urlopen):
        decision = {"decision": "allow", "units": 1, "remainingUnits": 99}
        resp = MagicMock()
        resp.read.return_value = json.dumps(decision).encode()
        resp.__enter__ = MagicMock(return_value=resp)
        resp.__exit__ = MagicMock(return_value=False)
        mock_urlopen.return_value = resp

        result = authorize(api_key="sk_test", token="tok_123", method="GET", path="/v1/data")
        assert result["decision"] == "allow"
        assert result["units"] == 1

    @patch("payfence.client.urlopen")
    def test_deny_response_via_http_error(self, mock_urlopen):
        decision = {"decision": "deny", "reason": "payment_required"}
        error = HTTPError(
            url="https://api.payfence.io/v1/authorize",
            code=402,
            msg="Payment Required",
            hdrs={},
            fp=BytesIO(json.dumps(decision).encode()),
        )
        mock_urlopen.side_effect = error

        result = authorize(api_key="sk_test", token="tok_123", method="POST", path="/v1/weather")
        assert result["decision"] == "deny"
        assert result["reason"] == "payment_required"

    @patch("payfence.client.urlopen")
    def test_custom_base_url(self, mock_urlopen):
        resp = MagicMock()
        resp.read.return_value = json.dumps({"decision": "allow"}).encode()
        resp.__enter__ = MagicMock(return_value=resp)
        resp.__exit__ = MagicMock(return_value=False)
        mock_urlopen.return_value = resp

        authorize(api_key="sk_test", token="tok_123", method="GET", path="/", base_url="https://custom.example.com/")

        # Verify the Request object URL
        req_arg = mock_urlopen.call_args[0][0]
        assert req_arg.full_url == "https://custom.example.com/v1/authorize"

    @patch("payfence.client.urlopen")
    def test_request_id_forwarded(self, mock_urlopen):
        resp = MagicMock()
        resp.read.return_value = json.dumps({"decision": "allow"}).encode()
        resp.__enter__ = MagicMock(return_value=resp)
        resp.__exit__ = MagicMock(return_value=False)
        mock_urlopen.return_value = resp

        authorize(api_key="sk_test", token="tok_123", method="GET", path="/", request_id="my-req-id")

        req_arg = mock_urlopen.call_args[0][0]
        body = json.loads(req_arg.data)
        assert body["requestId"] == "my-req-id"



import json
import logging
import uuid
from urllib.request import Request, urlopen
from urllib.error import URLError, HTTPError
from typing import Optional, Dict, Any

logger = logging.getLogger("payfence")


def authorize(
    api_key: str,
    token: str,
    method: str,
    path: str,
    request_id: Optional[str] = None,
    base_url: str = "https://api.payfence.io",
) -> Dict[str, Any]:
    """Call PayFence authorize endpoint. Returns decision dict."""
    url = f"{base_url.rstrip('/')}/v1/authorize"
    rid = request_id or str(uuid.uuid4())
    payload = json.dumps({
        "token": token,
        "method": method,
        "path": path,
        "requestId": rid,
    }).encode()

    req = Request(url, data=payload, method="POST")
    req.add_header("Content-Type", "application/json")
    req.add_header("Authorization", f"Bearer {api_key}")

    logger.debug("Authorizing: %s %s (requestId=%s)", method, path, rid)

    try:
        with urlopen(req, timeout=10) as resp:
            decision = json.loads(resp.read())
            logger.debug("Allow: %s %s", method, path)
            return decision
    except HTTPError as e:
        decision = json.loads(e.read())
        logger.info("Deny: %s %s — status=%d, reason=%s", method, path, e.code, decision.get("reason"))
        return decision
